//--------------------------------------------------------------------------
//
//  loadmap - an IDA plugin, which loads a VC++/BCC map file.
//  Copyright (C) 2000 Toshiyuki Tega. All rights reserved.
//
//  This software is provided 'as-is', without any express or implied
//  warranty.  In no event will the author be held liable for any damages
//  arising from the use of this software.
//
//--------------------------------------------------------------------------

#include  "mapread.h"

#include  <cstring>
#include  <cctype>

using namespace std;

namespace
{
  //  This is where the symbol table starts.
  const char  VC_SYMBOL_START[] =
    "  Address         Publics by Value              Rva+Base     Lib:Object";
  const char  BCC_SYMBOL_START[] =
    "  Address         Publics by Value";

  //  Used to skip any unnecessary VC++ map information.
  const char  VC_LINE_NUMBER[] = "Line numbers for ";
  const char  VC_FIXUP[]       = "FIXUPS: ";
  const char  VC_EXPORTS[]     = " Exports";

  //  Separators used when we cut tokens down.
  const char  TOKEN_SEP[] = " :\r\n";

  bool  is_hex(char c)
  {
    const int C = toupper(c);
    return isdigit(c) || ('A' <= C && C <= 'F');
  }

  int   get_digit(char c)
  {
    return isdigit(c) ? (c - '0') : (0xA + toupper(c) - 'A');
  }
}

//--------------------------------------------------------------------------
//  map_reader::map_reader()
//--------------------------------------------------------------------------
map_reader::map_reader(const char * map_name)
  : m_fp(fopen(map_name, "r")), m_pos(),
    m_state(SKIPPING_HEADER), m_type(MAP_FILE_TYPE_UNKNOWN)
{
}

//--------------------------------------------------------------------------
//  map_reader::~map_reader()
//--------------------------------------------------------------------------
map_reader::~map_reader()
{
}

//--------------------------------------------------------------------------
//  map_reader::get_hex()
//    Get next token and convert it to an integer if possible.
//--------------------------------------------------------------------------
bool  map_reader::get_hex(unsigned long * val, bool init)
{
  char *  top = init ? m_buf : 0;
  char *  tok = strtok(top, TOKEN_SEP);

  if(0 == tok)
    return false;

  *val = 0;
  for(size_t i = 0; i < strlen(tok); ++i)
  {
    if(false == is_hex(tok[i]))
      return false;
    *val <<= 4;
    *val += get_digit(tok[i]);
  }
  return true;
}

//--------------------------------------------------------------------------
//  map_reader::get_string()
//    Get next token.
//--------------------------------------------------------------------------
bool  map_reader::get_string(char ** str)
{
  return 0 != (*str = strtok(0, TOKEN_SEP));
}

//--------------------------------------------------------------------------
//  map_reader::get_line()
//    Read one line into the buffer m_buf.
//--------------------------------------------------------------------------
bool  map_reader::get_line()
{
  m_pos = 0;
  return 0 != fgets(m_buf, BUFSIZE, m_fp);
}

//--------------------------------------------------------------------------
//  map_reader::get_symbol()
//    Get segment index, segment offset, and symbol name.
//--------------------------------------------------------------------------
bool  map_reader::get_symbol(unsigned long * seg_no,
                             unsigned long * offset,
                             char **         name)
{
  return get_hex(seg_no, true) && get_hex(offset, false) && get_string(name);
}

//--------------------------------------------------------------------------
//  map_reader::skip_header()
//    Read and discard unused information until we get to the symbol table.
//--------------------------------------------------------------------------
bool  map_reader::skip_header()
{
  if(0 == m_fp)
    return true;

  while(SKIPPING_HEADER == m_state && false == eof())
  {
    if(get_line())
    {
      if(0 == strncmp(m_buf, VC_SYMBOL_START, sizeof(VC_SYMBOL_START) - 1))
      {
        m_type = MAP_FILE_TYPE_VC;
        m_state = LOADING_SYMBOL;
      }
      else if(0 == strncmp(m_buf, BCC_SYMBOL_START, sizeof(BCC_SYMBOL_START) - 1))
      {
        m_type = MAP_FILE_TYPE_BCC;
        m_state = LOADING_SYMBOL;
      }
    }
  }
  return LOADING_SYMBOL == m_state;
}

//--------------------------------------------------------------------------
//  map_reader::load_symbol()
//    Interpret a line and get symbol name etc. if possible.
//--------------------------------------------------------------------------
bool  map_reader::load_symbol(unsigned long * seg_no, unsigned long * offset, char ** name)
{
  if(0 == m_fp || LOADING_SYMBOL != m_state)
    return false;

  if(get_line())
  {
    //  Check the end of the symbol table.
    if(0 == strncmp(m_buf, VC_LINE_NUMBER, sizeof(VC_LINE_NUMBER) - 1) ||
       0 == strncmp(m_buf, VC_FIXUP,       sizeof(VC_FIXUP)       - 1) ||
       0 == strncmp(m_buf, VC_EXPORTS,     sizeof(VC_EXPORTS)     - 1))
      m_state = FINISHED;
    
    if(FINISHED != m_state)
    {
      //  14 means at least there's enough room for "XXXX:XXXXXXXX ".
      const int MINLEN = 14;
      if(strlen(m_buf) >= MINLEN)
        return get_symbol(seg_no, offset, name);
    }
  }
  return false;
}

//--------------------------------------------------------------------------
//  map_reader::eof()
//--------------------------------------------------------------------------
bool  map_reader::eof()
{
  return 0 == m_fp || feof(m_fp) || FINISHED == m_state;
}
